@@ -2,7 +2,7 @@
  * Modified from industrialio buffer test code, and Lenovo Yoga (2 Pro) orientation helper
  * Copyright (c) 2008 Jonathan Cameron
  * Copyright (c) 2014 Peter F. Patel-Schneider
- * Copyright (c) 2014 Bastien Nocera <hadess@hadess.net>
+ * Copyright (c) 2014, 2021 Bastien Nocera <hadess@hadess.net>
  * Copyright (c) 2015 Elad Alfassa <elad@fedoraproject.org>
  *
  * This program is free software; you can redistribute it and/or modify it
@@ -11,22 +11,33 @@
  */
 
 #include "iio-buffer-utils.h"
+#include "utils.h"
 
+#include <ctype.h>
 #include <fcntl.h>
+#include <float.h>
+#include <math.h>
 #include <string.h>
 #include <errno.h>
 #include <stdio.h>
 
+#define IIO_MIN_SAMPLING_FREQUENCY	10 /* Hz */
+
 /**
  * iio_channel_info - information about a given channel
  * @name: channel name
- * @scale: scale factor to be applied for conversion to si units
- * @offset: offset to be applied for conversion to si units
+ * @generic_name: the generic name of the channel
+ * @scale: scale factor to be applied for conversion to SI units
+ * @offset: offset to be applied for conversion to SI units
  * @index: the channel index in the buffer output
- * @bytes: number of bytes occupied in buffer output
+ * @is_signed: signed or unsigned
+ * @bits_used: number of valid bits of data
+ * @bytes: number of bytes occupied in buffer output (bits_used + padding)
+ * @shift: shift right by this before masking out bits_used
  * @mask: a bit mask for the raw output
- * @is_signed: is the raw value stored signed
+ * @be: little or big endian
  * @enabled: is this channel enabled
+ * @location: data offset for this channel inside the buffer (in bytes)
  **/
 struct iio_channel_info {
 	char *name;
@@ -34,12 +45,12 @@ struct iio_channel_info {
 	float scale;
 	float offset;
 	unsigned index;
-	unsigned bytes;
+	unsigned is_signed;
 	unsigned bits_used;
+	unsigned bytes;
 	unsigned shift;
 	guint64 mask;
 	unsigned be;
-	unsigned is_signed;
 	unsigned enabled;
 	unsigned location;
 };
@@ -47,7 +58,7 @@ struct iio_channel_info {
 static char *
 iioutils_break_up_name (const char *name)
 {
-	char **items, *ret;
+	g_auto(GStrv) items = NULL;
 	guint i;
 
 	items = g_strsplit (name, "_", -1);
@@ -58,10 +69,7 @@ iioutils_break_up_name (const char *name)
 		}
 	}
 
-	ret = g_strjoinv ("_", items);
-	g_strfreev (items);
-
-	return ret;
+	return g_strjoinv ("_", items);
 }
 
 /**
@@ -72,6 +80,8 @@ iioutils_break_up_name (const char *name)
  * @be: big endian
  * @device_dir: the iio device directory
  * @name: the channel name
+ *
+ * See `struct iio_chan_spec` in the kernel headers.
  **/
 static gboolean
 iioutils_get_type (unsigned   *is_signed,
@@ -85,29 +95,28 @@ iioutils_get_type (unsigned   *is_signed,
 		   const char *generic_name)
 {
 	int ret;
-	char *builtname;
-	char *filename;
+	g_autofree char *builtname = NULL;
+	g_autofree char *filename = NULL;
 	char signchar, endianchar;
 	unsigned padint;
-	FILE *sysfsfp;
+	g_autoptr(FILE) sysfsfp = NULL;
 
 	builtname = g_strdup_printf ("%s_type", name);
 	filename = g_build_filename (device_dir, "scan_elements", builtname, NULL);
-	g_free (builtname);
 
 	sysfsfp = fopen (filename, "r");
 	if (sysfsfp == NULL) {
+		g_clear_pointer (&builtname, g_free);
+		g_clear_pointer (&filename, g_free);
 		builtname = g_strdup_printf ("%s_type", generic_name);
 		filename = g_build_filename (device_dir, "scan_elements", builtname, NULL);
-		g_free (builtname);
 
 		sysfsfp = fopen (filename, "r");
-		if (sysfsfp == NULL) {
-			g_free (filename);
+		if (sysfsfp == NULL)
 			return FALSE;
-		}
 	}
 
+	/* See `iio_show_fixed_type()` in the IIO core for the format */
 	ret = fscanf (sysfsfp,
 		      "%ce:%c%u/%u>>%u",
 		      &endianchar,
@@ -115,30 +124,22 @@ iioutils_get_type (unsigned   *is_signed,
 		      bits_used,
 		      &padint, shift);
 
-	if (ret < 0) {
+	if (ret < 0 || ret != 5) {
 		g_warning ("Failed to pass scan type description for %s", filename);
-		fclose (sysfsfp);
-		g_free (filename);
 		return FALSE;
 	}
-	fclose (sysfsfp);
 
 	*be = (endianchar == 'b');
 	*bytes = padint / 8;
 	if (*bits_used == 64)
-		*mask = ~0;
+		*mask = ~G_GUINT64_CONSTANT(0);
 	else
-		*mask = (1 << *bits_used) - 1;
-	if (signchar == 's')
-		*is_signed = 1;
-	else
-		*is_signed = 0;
+		*mask = (G_GUINT64_CONSTANT(1) << *bits_used) - G_GUINT64_CONSTANT(1);
+	*is_signed = (signchar == 's');
 
 	g_debug ("Got type for %s: is signed: %d, bytes: %d, bits_used: %d, shift: %d, mask: 0x%" G_GUINT64_FORMAT ", be: %d",
 		 name, *is_signed, *bytes, *bits_used, *shift, *mask, *be);
 
-	g_free (filename);
-
 	return TRUE;
 }
 
@@ -149,40 +150,53 @@ iioutils_get_param_float (float      *output,
 			  const char *name,
 			  const char *generic_name)
 {
-	FILE *sysfsfp;
-	char *builtname, *filename;
-	int ret = 0;
+	g_autofree char *builtname = NULL;
+	g_autofree char *filename = NULL;
+	g_autofree char *contents = NULL;
+	g_autoptr(GError) error = NULL;
+
+	g_debug ("Trying to read '%s_%s' (name) from dir '%s'", name, param_name, device_dir);
 
 	builtname = g_strdup_printf ("%s_%s", name, param_name);
 	filename = g_build_filename (device_dir, builtname, NULL);
-	g_free (builtname);
 
-	sysfsfp = fopen (filename, "r");
-	if (sysfsfp) {
-		fscanf (sysfsfp, "%f", output);
-		fclose (sysfsfp);
-		g_free (filename);
-		return 0;
+	if (g_file_get_contents (filename, &contents, NULL, &error)) {
+		char *endptr;
+		*output = g_ascii_strtod (contents, &endptr);
+		if (*output != 0.0 || endptr != contents)
+			return 0;
+		g_warning ("Couldn't convert '%s' from %s to float", g_strchomp (contents), filename);
+	} else {
+		g_debug ("Failed to read float from %s: %s", filename, error->message);
+		g_clear_error (&error);
 	}
 
-	g_free (filename);
+	g_debug ("Trying to read '%s_%s' (generic name) from dir '%s'", generic_name, param_name, device_dir);
 
+	g_clear_pointer (&builtname, g_free);
+	g_clear_pointer (&filename, g_free);
 	builtname = g_strdup_printf ("%s_%s", generic_name, param_name);
 	filename = g_build_filename (device_dir, builtname, NULL);
-	g_free (builtname);
 
-	sysfsfp = fopen (filename, "r");
-	if (sysfsfp) {
-		fscanf (sysfsfp, "%f", output);
-		fclose (sysfsfp);
+	g_clear_pointer (&contents, g_free);
+	if (g_file_get_contents (filename, &contents, NULL, &error)) {
+		char *endptr;
+		*output = g_ascii_strtod (contents, &endptr);
+		if (*output == 0.0 && endptr == contents) {
+			g_warning ("Couldn't convert '%s' from %s to float", g_strchomp (contents), filename);
+			return -EINVAL;
+		}
 	} else {
-		ret = -errno;
-		g_warning ("Failed to read float from %s", filename);
+		if (g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
+			g_debug ("Failed to read float from non-existent %s", filename);
+			return -ENOENT;
+		} else {
+			g_warning ("Failed to read float from %s: %s", filename, error->message);
+			return -EINVAL;
+		}
 	}
 
-	g_free (filename);
-
-	return ret;
+	return 0;
 }
 
 static void
@@ -193,13 +207,23 @@ channel_info_free (iio_channel_info *ci)
 	g_free (ci);
 }
 
+static int
+compare_channel_index (gconstpointer a, gconstpointer b)
+{
+	const iio_channel_info *info_1 = *(iio_channel_info **) a;
+	const iio_channel_info *info_2 = *(iio_channel_info **) b;
+
+	return (int) (info_1->index - info_2->index);
+}
+
+G_DEFINE_AUTOPTR_CLEANUP_FUNC(iio_channel_info, channel_info_free)
+
 /* build_channel_array() - function to figure out what channels are present */
 static iio_channel_info **
 build_channel_array (const char        *device_dir,
 		     int               *counter)
 {
 	GDir *dp;
-	FILE *sysfsfp;
 	int ret;
 	const char *name;
 	char *scan_el_dir;
@@ -213,6 +237,7 @@ build_channel_array (const char        *device_dir,
 	dp = g_dir_open (scan_el_dir, 0, NULL);
 	if (dp == NULL) {
 		ret = -errno;
+		g_debug ("Could not open scan_elements dir '%s'", scan_el_dir);
 		g_free (scan_el_dir);
 		return NULL;
 	}
@@ -221,22 +246,22 @@ build_channel_array (const char        *device_dir,
 
 	while ((name = g_dir_read_name (dp)) != NULL) {
 		if (g_str_has_suffix (name, "_en")) {
-			char *filename, *index_name;
-			iio_channel_info *current;
+			g_autoptr(FILE) sysfsfp = NULL;
+			g_autofree char *filename = NULL;
+			g_autofree char *index_name = NULL;
+			g_autoptr(iio_channel_info) current = NULL;
 
 			filename = g_build_filename (scan_el_dir, name, NULL);
 			sysfsfp = fopen (filename, "r");
 			if (sysfsfp == NULL) {
-				g_free (filename);
+				g_debug ("Could not open scan_elements file '%s'", filename);
 				continue;
 			}
-			fscanf (sysfsfp, "%d", &ret);
-			fclose (sysfsfp);
-			if (!ret) {
-				g_free (filename);
+			if (fscanf (sysfsfp, "%d", &ret) != 1) {
+				g_debug ("Could not read from scan_elements file '%s'", filename);
 				continue;
 			}
-			g_free (filename);
+			g_clear_pointer (&sysfsfp, fclose);
 
 			current = g_new0 (iio_channel_info, 1);
 
@@ -245,17 +270,19 @@ build_channel_array (const char        *device_dir,
 			current->name = g_strndup (name, strlen(name) - strlen("_en"));
 			current->generic_name = iioutils_break_up_name (current->name);
 			if (g_strcmp0(current->generic_name, "in_rot_from_north_magnetic_tilt") == 0) {
-				current->generic_name = "in_rot";
+				current->generic_name = g_strdup ("in_rot");
 			}
 
 			index_name = g_strdup_printf ("%s_index", current->name);
+			g_clear_pointer (&filename, g_free);
 			filename = g_build_filename (scan_el_dir, index_name, NULL);
-			g_free (index_name);
 
 			sysfsfp = fopen (filename, "r");
-			fscanf (sysfsfp, "%u", &current->index);
-			fclose (sysfsfp);
-			g_free (filename);
+			if (sysfsfp == NULL)
+				goto error;
+			ret = fscanf (sysfsfp, "%u", &current->index);
+			if (ret != 1)
+				goto error;
 
 			/* Find the scale */
 			ret = iioutils_get_param_float (&current->scale,
@@ -263,7 +290,9 @@ build_channel_array (const char        *device_dir,
 							device_dir,
 							current->name,
 							current->generic_name);
-			if (ret < 0)
+			if ((ret < 0) && (ret != -ENOENT))
+				goto error;
+			if (current->scale == 0.0)
 				goto error;
 
 			ret = iioutils_get_param_float (&current->offset,
@@ -271,7 +300,7 @@ build_channel_array (const char        *device_dir,
 							device_dir,
 							current->name,
 							current->generic_name);
-			if (ret < 0)
+			if ((ret < 0) && (ret != -ENOENT))
 				goto error;
 
 			ret = iioutils_get_type (&current->is_signed,
@@ -288,21 +317,23 @@ build_channel_array (const char        *device_dir,
 				g_warning ("Could not parse name %s, generic name %s",
 					   current->name, current->generic_name);
 			} else {
-				g_ptr_array_add (array, current);
+				g_ptr_array_add (array, g_steal_pointer (&current));
 			}
 		}
 	}
 	g_dir_close (dp);
 	g_free (scan_el_dir);
 
+	g_ptr_array_sort (array, compare_channel_index);
+
 	*counter = array->len;
 	ret_array = (iio_channel_info **) g_ptr_array_free (array, FALSE);
 
 	for (i = 0; i < *counter; i++) {
 		iio_channel_info *ci = ret_array[i];
 
-		g_debug ("Built channel array for %s: is signed: %d, bytes: %d, bits_used: %d, shift: %d, mask: 0x%" G_GUINT64_FORMAT ", be: %d",
-			 ci->name, ci->is_signed, ci->bytes, ci->bits_used, ci->shift, ci->mask, ci->be);
+		g_debug ("Built channel array for %s: index: %d, is signed: %d, bytes: %d, bits_used: %d, shift: %d, mask: 0x%" G_GUINT64_FORMAT ", be: %d",
+			 ci->name, ci->index, ci->is_signed, ci->bytes, ci->bits_used, ci->shift, ci->mask, ci->be);
 	}
 
 	return ret_array;
@@ -323,7 +354,7 @@ _write_sysfs_int (const char *filename,
 		  int         val2)
 {
 	int ret = 0;
-	FILE *sysfsfp;
+	g_autoptr(FILE) sysfsfp = NULL;
 	int test;
 	char *temp;
 	temp = g_build_filename (basedir, filename, NULL);
@@ -337,8 +368,8 @@ _write_sysfs_int (const char *filename,
 		fprintf(sysfsfp, "%d %d", val, val2);
 	else
 		fprintf(sysfsfp, "%d", val);
+	g_clear_pointer (&sysfsfp, fclose);
 
-	fclose(sysfsfp);
 	if (verify) {
 		sysfsfp = fopen(temp, "r");
 		if (sysfsfp == NULL) {
@@ -346,13 +377,12 @@ _write_sysfs_int (const char *filename,
 			ret = -errno;
 			goto error_free;
 		}
-		fscanf(sysfsfp, "%d", &test);
-		if (test != val) {
+		if (fscanf(sysfsfp, "%d", &test) != 1 ||
+		    test != val) {
 			g_warning ("Possible failure in int write %d to %s",
 				   val, temp);
 			ret = -1;
 		}
-		fclose(sysfsfp);
 	}
 error_free:
 	g_free (temp);
@@ -373,39 +403,29 @@ _write_sysfs_string (const char *filename,
 		     const char *val,
 		     int         verify)
 {
-	int ret = 0;
-	FILE *sysfsfp;
-	char *temp;
+	g_autoptr(FILE) sysfsfp = NULL;
+	g_autofree char *temp = NULL;
+	g_autofree char *contents = NULL;
 
 	temp = g_build_filename (basedir, filename, NULL);
 	sysfsfp = fopen (temp, "w");
 	if (sysfsfp == NULL) {
-		ret = -errno;
-		goto error_free;
+		return -errno;
 	}
 	fprintf(sysfsfp, "%s", val);
-	fclose(sysfsfp);
 
 	/* Verify? */
 	if (!verify)
-		goto error_free;
-	sysfsfp = fopen(temp, "r");
-	if (sysfsfp == NULL) {
-		ret = -errno;
-		goto error_free;
+		return 0;
+	g_clear_pointer (&sysfsfp, fclose);
+	if (!g_file_get_contents (temp, &contents, NULL, NULL) ||
+	    g_strcmp0 (g_strchomp (contents), val) != 0) {
+		g_warning ("Possible failure in string write of '%s' Should be '%s' written to %s\n",
+			   contents ?: "(null)", val, temp);
+		return -1;
 	}
-	fscanf(sysfsfp, "%s", temp);
-	if (strcmp(temp, val) != 0) {
-		g_warning ("Possible failure in string write of %s Should be %s written to %s\\%s\n",
-			   temp, val, basedir, filename);
-		ret = -1;
-	}
-	fclose(sysfsfp);
 
-error_free:
-	g_free(temp);
-
-	return ret;
+	return 0;
 }
 
 /**
@@ -449,6 +469,25 @@ size_from_channelarray (iio_channel_info **channels,
 	return bytes;
 }
 
+#define GUINT8_FROM_BE(x) (x)
+#define GUINT8_FROM_LE(x) (x)
+#define PROCESS_CHANNEL_BITS(bits)							\
+{											\
+	guint##bits input;								\
+											\
+	input = *(guint##bits *)(data + info->location);				\
+	input = info->be ? GUINT##bits##_FROM_BE (input) : GUINT##bits##_FROM_LE (input);\
+	input >>= info->shift;								\
+	input &= info->mask;								\
+	if (info->is_signed) {								\
+		gint##bits val = (gint##bits)(input << (bits - info->bits_used)) >> (bits - info->bits_used);	\
+		val += info->offset;							\
+		*ch_val = val;								\
+	} else {									\
+		*ch_val = input + info->offset;						\
+	}										\
+}
+
 /**
  * process_scan_1() - get an integer value for a particular channel
  * @data:               pointer to the start of the scan
@@ -460,52 +499,165 @@ size_from_channelarray (iio_channel_info **channels,
  **/
 void
 process_scan_1 (char              *data,
-    BufferDrvData        *buffer_data,
-		char              *ch_name,
+		BufferDrvData     *buffer_data,
+		const char        *ch_name,
 		int               *ch_val,
 		gdouble           *ch_scale,
 		gboolean          *ch_present)
 {
 	int k;
+
+	*ch_present = FALSE;
+
 	for (k = 0; k < buffer_data->channels_count; k++) {
-		if (strcmp (buffer_data->channels[k]->name, ch_name) != 0)
+		struct iio_channel_info *info = buffer_data->channels[k];
+
+		if (strcmp (info->name, ch_name) != 0)
 			continue;
 
-		switch (buffer_data->channels[k]->bytes) {
-			/* only a few cases implemented so far */
-		case 4:
-			if (!buffer_data->channels[k]->is_signed) {
-				guint32 val = *(guint32 *) (data + buffer_data->channels[k]->location);
-				val = val >> buffer_data->channels[k]->shift;
-				if (buffer_data->channels[k]->bits_used < 32) val &= ((guint32) 1 << buffer_data->channels[k]->bits_used) - 1;
-				*ch_val = (int) val;
-				*ch_present = TRUE;
-			} else {
-				gint32 val = *(gint32 *) (data + buffer_data->channels[k]->location);
-				val = val >> buffer_data->channels[k]->shift;
-				if (buffer_data->channels[k]->bits_used < 32) val &= ((guint32) 1 << buffer_data->channels[k]->bits_used) - 1;
-				val = (gint32) (val << (32 - buffer_data->channels[k]->bits_used)) >> (32 - buffer_data->channels[k]->bits_used);
-				*ch_val = (int) val;
-				if (buffer_data->channels[k]->scale)
-					*ch_scale = buffer_data->channels[k]->scale;
-				else
-					*ch_scale = 1.0;
-				*ch_present = TRUE;
-			}
+		g_debug ("process_scan_1: channel_index: %d, chan_name: %s, channel_data_index: %d location: %d bytes: %d is_signed: %d be: %d shift: %d bits_used: %d",
+			 k, info->name, info->index, info->location,
+			 info->bytes, info->is_signed, info->be,
+			 info->shift, info->bits_used);
+
+		switch (info->bytes) {
+		case 1:
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wduplicated-branches"
+			PROCESS_CHANNEL_BITS(8);
+#pragma GCC diagnostic pop
 			break;
 		case 2:
+			PROCESS_CHANNEL_BITS(16);
+			break;
+		case 4:
+			PROCESS_CHANNEL_BITS(32);
+			break;
 		case 8:
-			g_error ("Process %d bytes channels not supported yet", buffer_data->channels[k]->bytes);
+			PROCESS_CHANNEL_BITS(64);
+			break;
 		default:
-			g_assert_not_reached ();
+			g_error ("Process %d bytes channels not supported", info->bytes);
 			break;
 		}
+
+		*ch_scale = info->scale;
+		*ch_present = TRUE;
+		break;
 	}
 
 	if (!*ch_present)
 		g_warning ("IIO channel '%s' could not be found", ch_name);
 }
 
+/**
+ * iio_round_sampling_frequency: Round sampling_frequency to a supported value
+ * @dev: the IIO device to round the sampling frequency for
+ * @name: name of the *sampling_frequency attribute for which to round value
+ * @desired: desired value
+ *
+ * Check "<name>_available" and if that attribute exists round the desired
+ * value to a supported value, rounding up where possible.
+ */
+static double
+iio_round_sampling_frequency (GUdevDevice *dev, const char *name, double desired)
+{
+	g_autofree char *available_attr = NULL;
+	const char *available_str;
+	char *endptr;
+	double closest_higher = DBL_MAX;
+	double closest_lower = 0.0;
+	double available;
+
+	available_attr = g_strconcat (name, "_available", NULL);
+	available_str = g_udev_device_get_sysfs_attr (dev, available_attr);
+
+	if (!available_str)
+		return desired;
+
+	while (*available_str) {
+		available = g_ascii_strtod (available_str, &endptr);
+		if (available_str == endptr || (*endptr && !isspace (*endptr)))
+			break;
+
+		available_str = endptr;
+
+		/* 0 disables sampling, skip */
+		if (available == 0.0)
+			continue;
+
+		if (available >= desired) {
+			if (available < closest_higher)
+				closest_higher = available;
+		} else {
+			if (available > closest_lower)
+				closest_lower = available;
+		}
+	}
+
+	/* Prefer higher values */
+	if (closest_higher != DBL_MAX)
+		return closest_higher;
+
+	if (closest_lower != 0.0)
+		return closest_lower;
+
+	return desired;
+}
+
+/**
+ * iio_fixup_sampling_frequency: Fixup devices *sampling_frequency attributes
+ * @dev: the IIO device to fix the sampling frequencies for
+ *
+ * Make sure devices with *sampling_frequency attributes are sampling at
+ * 10Hz or more. This fixes 2 problems:
+ * 1) Some buffered devices default their sampling_frequency to 0Hz and then
+ * never produce any readings.
+ * 2) Some polled devices default to 1Hz and wait for a fresh sample before
+ * returning from sysfs *_raw reads, blocking all of iio-sensor-proxy for
+ * multiple seconds
+ **/
+gboolean
+iio_fixup_sampling_frequency (GUdevDevice *dev)
+{
+	GDir *dir;
+	const char *device_dir;
+	const char *name;
+	g_autoptr(GError) error = NULL;
+	char sample_freq_str[G_ASCII_DTOSTR_BUF_SIZE];
+	double sample_freq;
+
+	device_dir = g_udev_device_get_sysfs_path (dev);
+	dir = g_dir_open (g_udev_device_get_sysfs_path (dev), 0, &error);
+	if (!dir) {
+		g_warning ("Failed to open directory '%s': %s", device_dir, error->message);
+		return FALSE;
+	}
+
+	while ((name = g_dir_read_name (dir))) {
+		if (g_str_has_suffix (name, "sampling_frequency") == FALSE)
+			continue;
+
+		sample_freq = g_udev_device_get_sysfs_attr_as_double (dev, name);
+		if (sample_freq >= IIO_MIN_SAMPLING_FREQUENCY)
+			continue; /* Continue with pre-set sample freq. */
+
+		/* Sample freq too low, set it to a supported freq close to 10Hz */
+		sample_freq = iio_round_sampling_frequency (dev, name, IIO_MIN_SAMPLING_FREQUENCY);
+
+		/*
+		 * Some sample_freq IIO sysfs attr only accept integers, most will
+		 * accept floats but not all. g_ascii_dtostr() uses the shortest
+		 * possible representation, omitting the "." for whole numbers.
+		 */
+		g_ascii_dtostr (sample_freq_str, G_ASCII_DTOSTR_BUF_SIZE, sample_freq);
+		if (write_sysfs_string (name, device_dir, sample_freq_str) < 0)
+			g_warning ("Could not fix sample-freq for %s/%s", device_dir, name);
+	}
+	g_dir_close (dir);
+	return TRUE;
+}
+
 /**
  * enable_sensors: enable all the sensors in a device
  * @device_dir: the IIO device directory in sysfs
@@ -518,11 +670,13 @@ enable_sensors (GUdevDevice *dev,
 	GDir *dir;
 	char *device_dir;
 	const char *name;
-	gboolean ret = TRUE;
+	gboolean ret = FALSE;
+	g_autoptr(GError) error = NULL;
 
 	device_dir = g_build_filename (g_udev_device_get_sysfs_path (dev), "scan_elements", NULL);
-	dir = g_dir_open (device_dir, 0, NULL);
+	dir = g_dir_open (device_dir, 0, &error);
 	if (!dir) {
+		g_warning ("Failed to open directory '%s': %s", device_dir, error->message);
 		g_free (device_dir);
 		return FALSE;
 	}
@@ -536,6 +690,8 @@ enable_sensors (GUdevDevice *dev,
 		/* Already enabled? */
 		path = g_strdup_printf ("scan_elements/%s", name);
 		if (g_udev_device_get_sysfs_attr_as_boolean (dev, path)) {
+			g_debug ("Already enabled sensor %s/%s", device_dir, name);
+			ret = TRUE;
 			g_free (path);
 			continue;
 		}
@@ -544,15 +700,20 @@ enable_sensors (GUdevDevice *dev,
 		/* Enable */
 		if (write_sysfs_int (name, device_dir, enable) < 0) {
 			g_warning ("Could not enable sensor %s/%s", device_dir, name);
-			ret = FALSE;
 			continue;
 		}
 
+		ret = TRUE;
 		g_debug ("Enabled sensor %s/%s", device_dir, name);
 	}
 	g_dir_close (dir);
 	g_free (device_dir);
 
+	if (!ret) {
+		g_warning ("Failed to enable any sensors for device '%s'",
+			   g_udev_device_get_sysfs_path (dev));
+	}
+
 	return ret;
 }
 
@@ -563,12 +724,14 @@ enable_ring_buffer (BufferDrvData *data)
 
 	/* Setup ring buffer parameters */
 	ret = write_sysfs_int("buffer/length", data->dev_dir_name, 128);
-	if (ret < 0)
+	if (ret < 0) {
+		g_warning ("Failed to set ring buffer length for %s", data->dev_dir_name);
 		return FALSE;
+	}
 	/* Enable the buffer */
 	ret = write_sysfs_int_and_verify("buffer/enable", data->dev_dir_name, 1);
 	if (ret < 0) {
-		printf("Unable to enable the buffer %d\n", ret);
+		g_warning ("Unable to enable ring buffer for %s", data->dev_dir_name);
 		return FALSE;
 	}
 
@@ -632,6 +795,7 @@ buffer_drv_data_free (BufferDrvData *buffer_data)
 	for (i = 0; i < buffer_data->channels_count; i++)
 		channel_info_free (buffer_data->channels[i]);
 	g_free (buffer_data->channels);
+	g_free (buffer_data);
 }
 
 BufferDrvData *
@@ -645,7 +809,8 @@ buffer_drv_data_new (GUdevDevice *device,
 	buffer_data->trigger_name = g_strdup (trigger_name);
 	buffer_data->device = g_object_ref (device);
 
-	if (!enable_sensors (device, 1) ||
+	if (!iio_fixup_sampling_frequency (device) ||
+	    !enable_sensors (device, 1) ||
 	    !enable_trigger (buffer_data) ||
 	    !enable_ring_buffer (buffer_data) ||
 	    !build_channels (buffer_data)) {
@@ -656,3 +821,23 @@ buffer_drv_data_new (GUdevDevice *device,
 	return buffer_data;
 }
 
+IIOSensorData *
+iio_sensor_data_new (gsize size)
+{
+	IIOSensorData *data;
+
+	g_return_val_if_fail (size != 0, NULL);
+
+	data = g_new0 (IIOSensorData, 1);
+	data->data = g_malloc0 (size);
+	return data;
+}
+
+void
+iio_sensor_data_free (IIOSensorData *data)
+{
+	if (data == NULL)
+		return;
+	g_free (data->data);
+	g_free (data);
+}
